package com.ghostsq.commander.sftp;

import android.content.Context;
import android.support.test.InstrumentationRegistry;
import android.util.Log;

import com.jcraft.jsch.Channel;
import com.jcraft.jsch.ChannelSftp;
import com.jcraft.jsch.JSch;
import com.jcraft.jsch.Logger;
import com.jcraft.jsch.Session;
import com.jcraft.jsch.UIKeyboardInteractive;
import com.jcraft.jsch.UserInfo;

import org.junit.Test;

import java.io.File;

import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

public class ConnectionTest {
    final static String TAG = "JSch.test";

    class MyUserInfo implements UserInfo, UIKeyboardInteractive {
        @Override
        public String getPassphrase() {
            Log.d( TAG, "Wants Passphrase" );
            return null;
        }
        @Override
        public String getPassword() {
            Log.d( TAG, "Wants Password" );
            return "password";
        }
        @Override
        public boolean promptPassword( String message ) {
            Log.d( TAG, "prompts Password: " + message );
            return true;
        }
        @Override
        public boolean promptPassphrase( String message ) {
            Log.d( TAG, "prompts Passphrase: " + message );
            return true;
        }
        @Override
        public boolean promptYesNo( String message ) {
            Log.d( TAG, "prompts Y/N: " + message );
            return true;
        }
        @Override
        public void showMessage( String message ) {
            Log.d( TAG, "Message: " + message );
        }

        @Override
        public String[] promptKeyboardInteractive( String destination, String name, String instruction, String[] prompt, boolean[] echo ) {
            Log.d( TAG, "destination: " + destination );
            Log.d( TAG, "name: " + name );
            Log.d( TAG, "instruction: " + instruction );
            for( String pr : prompt )
                Log.d( TAG, "prompt: " + pr );
            return new String[] { "PASSWORD" };
        }
    }

    @Test
    public void doTest() {
        new Client().start();
    }

    class Client extends Thread {

        @Override
        public void run() {
            connect();
        }

        private void connect() {
            try {
                JSch jsch = new JSch();

                JSch.setLogger( new Logger() {
                    @Override
                    public boolean isEnabled( int level ) {
                        return true;
                    }
                    @Override
                    public void log( int level, String message ) {
                        int all;
                        switch( level ) {
                            case Logger.INFO:  all = Log.INFO;  break;
                            case Logger.WARN:  all = Log.WARN;  break;
                            case Logger.ERROR:
                            case Logger.FATAL: all = Log.ERROR; break;
                            case Logger.DEBUG:
                            default:
                                all = Log.DEBUG;
                        }
                        Log.println( all, "JSCH", message );
                    }
                } );


//                Session session = jsch.getSession( "demo", "test.rebex.net", 22 );
                Session session = jsch.getSession( "username", "192.168.1.250", 22 );

                UserInfo ui = new MyUserInfo();
                session.setUserInfo(ui);
                session.connect();
                Log.d( TAG, "Session connected. " + session.toString() );
                Channel channel = session.openChannel("sftp");
                channel.connect();
                ChannelSftp c = (ChannelSftp)channel;
                Log.d( TAG, "ChannelSftp connected. " + c.toString() );
                java.util.Vector<ChannelSftp.LsEntry> vv = c.ls( "/" );
                assertNotNull( "No directory list returned!!!", vv );
                assertTrue( "Returned list is empty!!!", vv.size() > 0 );
                for( int i = 0; i < vv.size(); i++ ) {
                    ChannelSftp.LsEntry entry = vv.elementAt( i );
                    Log.v( TAG, entry.getLongname() );
                }
                Context appContext = InstrumentationRegistry.getInstrumentation().getTargetContext();
                File dest = appContext.getExternalFilesDir( null );
                c.get( "/readme.txt", dest.getAbsolutePath() );
                assertTrue( new File( dest, "readme.txt" ).exists() );

            } catch( Exception e ) {
                fail(e.getMessage());
            }
        }
    }
}
